@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtBluetooth module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include <QtCore/QGlobalStatic>
 #include <QtCore/QLoggingCategory>
@@ -47,58 +11,56 @@
 #include "objectmanager_p.h"
 #include "properties_p.h"
 #include "adapter1_bluez5_p.h"
-#include "manager_p.h"
 
 QT_BEGIN_NAMESPACE
 
+QT_IMPL_METATYPE_EXTERN(InterfaceList)
+QT_IMPL_METATYPE_EXTERN(ManufacturerDataList)
+QT_IMPL_METATYPE_EXTERN(ServiceDataList)
+QT_IMPL_METATYPE_EXTERN(ManagedObjectList)
+
+
 Q_DECLARE_LOGGING_CATEGORY(QT_BT_BLUEZ)
 
+using namespace QtBluetoothPrivate; // for D-Bus wrappers
+
 typedef enum Bluez5TestResultType
 {
     BluezVersionUnknown,
-    BluezVersion4,
     BluezVersion5,
     BluezNotAvailable
 } Bluez5TestResult;
 
-Q_GLOBAL_STATIC_WITH_ARGS(Bluez5TestResult, bluezVersion, (BluezVersionUnknown));
-Q_GLOBAL_STATIC_WITH_ARGS(QVersionNumber, bluezDaemonVersion, (QVersionNumber()));
+Q_GLOBAL_STATIC_WITH_ARGS(Bluez5TestResult, bluezVersion, (BluezVersionUnknown))
+Q_GLOBAL_STATIC_WITH_ARGS(QVersionNumber, bluezDaemonVersion, (QVersionNumber()))
 
-bool isBluez5()
+/*
+    Ensures that the DBus types are registered
+ */
+
+void initializeBluez5()
 {
     if (*bluezVersion() == BluezVersionUnknown) {
         OrgFreedesktopDBusObjectManagerInterface manager(QStringLiteral("org.bluez"),
                                                          QStringLiteral("/"),
                                                          QDBusConnection::systemBus());
-
         qDBusRegisterMetaType<InterfaceList>();
         qDBusRegisterMetaType<ManagedObjectList>();
         qDBusRegisterMetaType<ManufacturerDataList>();
+        qDBusRegisterMetaType<ServiceDataList>();
 
         QDBusPendingReply<ManagedObjectList> reply = manager.GetManagedObjects();
         reply.waitForFinished();
         if (reply.isError()) {
-            // not Bluez 5.x
-            OrgBluezManagerInterface manager_bluez4(QStringLiteral("org.bluez"),
-                                             QStringLiteral("/"),
-                                             QDBusConnection::systemBus());
-            QDBusPendingReply<QList<QDBusObjectPath> > reply
-                    = manager_bluez4.ListAdapters();
-            reply.waitForFinished();
-            if (reply.isError()) {
-                *bluezVersion() = BluezNotAvailable;
-                qWarning() << "Cannot find a running Bluez. Please check the Bluez installation.";
-            } else {
-                *bluezVersion() = BluezVersion4;
-                qCDebug(QT_BT_BLUEZ) << "Bluez 4 detected.";
-            }
+            *bluezVersion() = BluezNotAvailable;
+            qWarning() << "Cannot find a compatible running Bluez. "
+                          "Please check the Bluez installation. "
+                          "QtBluetooth requires at least BlueZ version 5.";
         } else {
             *bluezVersion() = BluezVersion5;
             qCDebug(QT_BT_BLUEZ) << "Bluez 5 detected.";
         }
     }
-
-    return (*bluezVersion() == BluezVersion5);
 }
 
 /*
@@ -177,7 +139,7 @@ bool mandatoryHciIoctlsAvailable()
     return true;
 }
 
-/*!
+/*
  * This function returns the version of bluetoothd in use on the system.
  * This is required to determine which QLEControllerPrivate implementation
  * is required. The following version tags are of significance:
@@ -195,6 +157,14 @@ bool mandatoryHciIoctlsAvailable()
 QVersionNumber bluetoothdVersion()
 {
     if (bluezDaemonVersion()->isNull()) {
+        // Register DBus specific  meta types (copied from isBluez5())
+        // Not all code paths run through isBluez5() but still need the
+        // registration.
+        qDBusRegisterMetaType<InterfaceList>();
+        qDBusRegisterMetaType<ManagedObjectList>();
+        qDBusRegisterMetaType<ManufacturerDataList>();
+        qDBusRegisterMetaType<ServiceDataList>();
+
         qCDebug(QT_BT_BLUEZ) << "Detecting bluetoothd version";
         //Order of matching
         // 1. Pick whatever the user decides via BLUETOOTH_FORCE_DBUS_LE_VERSION
@@ -351,11 +321,7 @@ QtBluezDiscoveryManager::~QtBluezDiscoveryManager()
 
 QtBluezDiscoveryManager *QtBluezDiscoveryManager::instance()
 {
-    if (isBluez5())
-        return discoveryManager();
-
-    Q_ASSERT(false);
-    return nullptr;
+    return discoveryManager();
 }
 
 bool QtBluezDiscoveryManager::registerDiscoveryInterest(const QString &adapterPath)
@@ -373,8 +339,8 @@ bool QtBluezDiscoveryManager::registerDiscoveryInterest(const QString &adapterPa
 
     OrgFreedesktopDBusPropertiesInterface *propIface = new OrgFreedesktopDBusPropertiesInterface(
                 QStringLiteral("org.bluez"), adapterPath, QDBusConnection::systemBus());
-    connect(propIface, SIGNAL(PropertiesChanged(QString,QVariantMap,QStringList)),
-            SLOT(PropertiesChanged(QString,QVariantMap,QStringList)));
+    connect(propIface, &OrgFreedesktopDBusPropertiesInterface::PropertiesChanged,
+            this, &QtBluezDiscoveryManager::PropertiesChanged);
     data->propteryListener = propIface;
 
     OrgBluezAdapter1Interface iface(QStringLiteral("org.bluez"), adapterPath,
@@ -437,7 +403,8 @@ void QtBluezDiscoveryManager::InterfacesRemoved(const QDBusObjectPath &object_pa
 
 void QtBluezDiscoveryManager::PropertiesChanged(const QString &interface,
                                                 const QVariantMap &changed_properties,
-                                                const QStringList &invalidated_properties)
+                                                const QStringList &invalidated_properties,
+                                                const QDBusMessage &)
 {
     Q_UNUSED(invalidated_properties);
 
@@ -482,11 +449,13 @@ void QtBluezDiscoveryManager::removeAdapterFromMonitoring(const QString &dbusPat
     emit discoveryInterrupted(dbusPath);
 }
 
-/*!
-    Finds the path for the local adapter with \a wantedAddress or an empty string
-    if no local adapter with the given address can be found.
-    If \a wantedAddress is \c null it returns the first/default adapter or an empty
-    string if none is available.
+/*
+    Finds the path for the local adapter with \a wantedAddress or returns an
+    empty string if no local adapter with the given address can be found.
+
+    If \a wantedAddress is \c null it returns the first powered-on adapter, or
+    the first adapter if all of them are powered off, or an empty string if
+    none is available.
 
     If \a ok is false the lookup was aborted due to a dbus error and this function
     returns an empty string.
@@ -506,8 +475,13 @@ QString findAdapterForAddress(const QBluetoothAddress &wantedAddress, bool *ok =
         return QString();
     }
 
-    typedef QPair<QString, QBluetoothAddress> AddressForPathType;
-    QList<AddressForPathType> localAdapters;
+    struct AdapterInfo
+    {
+        QString path;
+        QBluetoothAddress address;
+        bool powered;
+    };
+    QList<AdapterInfo> localAdapters;
 
     ManagedObjectList managedObjectList = reply.value();
     for (ManagedObjectList::const_iterator it = managedObjectList.constBegin(); it != managedObjectList.constEnd(); ++it) {
@@ -518,12 +492,13 @@ QString findAdapterForAddress(const QBluetoothAddress &wantedAddress, bool *ok =
             const QString &iface = jt.key();
 
             if (iface == QStringLiteral("org.bluez.Adapter1")) {
-                AddressForPathType pair;
-                pair.first = path.path();
-                pair.second = QBluetoothAddress(ifaceList.value(iface).value(
-                                          QStringLiteral("Address")).toString());
-                if (!pair.second.isNull())
-                    localAdapters.append(pair);
+                AdapterInfo info;
+                info.path = path.path();
+                info.address = QBluetoothAddress(ifaceList.value(iface).value(
+                                                    QStringLiteral("Address")).toString());
+                info.powered = ifaceList.value(iface).value(QStringLiteral("Powered")).toBool();
+                if (!info.address.isNull())
+                    localAdapters.append(info);
                 break;
             }
         }
@@ -535,17 +510,66 @@ QString findAdapterForAddress(const QBluetoothAddress &wantedAddress, bool *ok =
     if (localAdapters.isEmpty())
         return QString(); // -> no local adapter found
 
-    if (wantedAddress.isNull())
-        return localAdapters.front().first; // -> return first found adapter
+    auto findFirstPowered = [&localAdapters]() {
+        Q_ASSERT(!localAdapters.isEmpty());
+        auto it = std::find_if(localAdapters.cbegin(), localAdapters.cend(),
+                               [](const AdapterInfo &info) {
+                                   return info.powered == true;
+                               });
+        if (it != localAdapters.cend())
+            return it->path;
+        // simply return first if none is powered
+        return localAdapters.cbegin()->path;
+    };
 
-    for (const AddressForPathType &pair : qAsConst(localAdapters)) {
-        if (pair.second == wantedAddress)
-            return pair.first; // -> found local adapter with wanted address
+    if (wantedAddress.isNull())
+        return findFirstPowered();
+
+    for (const AdapterInfo &info : std::as_const(localAdapters)) {
+        if (info.address == wantedAddress)
+            return info.path; // -> found local adapter with wanted address
     }
 
     return QString(); // nothing matching found
 }
 
+/*
+
+  Checks the presence of peripheral interface and returns path to the adapter
+
+*/
+
+QString adapterWithDBusPeripheralInterface(const QBluetoothAddress &localAddress)
+{
+    initializeBluez5();
+
+    // First find the object path to the desired adapter
+    bool ok = false;
+    const QString hostAdapterPath = findAdapterForAddress(localAddress, &ok);
+    if (!ok || hostAdapterPath.isEmpty())
+        return {};
+
+    // Then check if that adapter provides peripheral dbus interface
+    OrgFreedesktopDBusObjectManagerInterface manager(QStringLiteral("org.bluez"),
+                                                     QStringLiteral("/"),
+                                                     QDBusConnection::systemBus());
+    QDBusPendingReply<ManagedObjectList> reply = manager.GetManagedObjects();
+    reply.waitForFinished();
+    if (reply.isError())
+        return {};
+
+    using namespace Qt::StringLiterals;
+    // For example /org/bluez/hci0 contains org.bluezLEAdvertisingManager1
+    const bool peripheralSupported = reply.value()
+                   .value(QDBusObjectPath(hostAdapterPath))
+                   .contains("org.bluez.LEAdvertisingManager1"_L1);
+
+    qCDebug(QT_BT_BLUEZ) << "Peripheral role"
+                         << (peripheralSupported ? "" : "not")
+                         << "supported on" << hostAdapterPath;
+    return peripheralSupported ? hostAdapterPath : QString{};
+}
+
 /*
     Removes every character that cannot be used in QDbusObjectPath
 
@@ -554,7 +578,7 @@ QString findAdapterForAddress(const QBluetoothAddress &wantedAddress, bool *ok =
 QString sanitizeNameForDBus(const QString &text)
 {
     QString appName = text;
-    for (int i = 0; i < appName.length(); i++) {
+    for (qsizetype i = 0; i < appName.size(); ++i) {
         ushort us = appName[i].unicode();
         bool valid = (us >= 'a' && us <= 'z')
                       || (us >= 'A' && us <= 'Z')
@@ -569,3 +593,5 @@ QString sanitizeNameForDBus(const QString &text)
 }
 
 QT_END_NAMESPACE
+
+#include "moc_bluez5_helper_p.cpp"
