@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 Centria research and development
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtNfc module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 Centria research and development
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qnearfieldtarget_android_p.h"
 #include "android/androidjninfc_p.h"
@@ -111,7 +75,7 @@ bool QNearFieldTargetPrivateImpl::disconnect()
     bool connected = tagTech.callMethod<jboolean>("isConnected");
     if (!connected)
         return false;
-    auto methodId = env.findMethod(tagTech.objectClass(), "close", "()V");
+    auto methodId = env.findMethod<void>(tagTech.objectClass(), "close");
     if (!methodId)
         return false;
     env->CallVoidMethod(tagTech.object(), methodId);
@@ -149,15 +113,13 @@ QNearFieldTarget::RequestId QNearFieldTargetPrivateImpl::readNdefMessages()
     }
 
     // Get NdefMessage object
-    QJniObject ndefMessage = tagTech.callObjectMethod("getNdefMessage", "()Landroid/nfc/NdefMessage;");
+    QJniObject ndefMessage = tagTech.callMethod<QtJniTypes::NdefMessage>("getNdefMessage");
     if (!ndefMessage.isValid()) {
         reportError(QNearFieldTarget::NdefReadError, requestId);
         return requestId;
     }
 
-    // Convert to byte array
-    QJniObject ndefMessageBA = ndefMessage.callObjectMethod("toByteArray", "()[B");
-    QByteArray ndefMessageQBA = jbyteArrayToQByteArray(ndefMessageBA.object<jbyteArray>());
+    QByteArray ndefMessageQBA = ndefMessage.callMethod<jbyte[]>("toByteArray").toContainer();
 
     // Sending QNdefMessage, requestCompleted and exit.
     QNdefMessage qNdefMessage = QNdefMessage::fromByteArray(ndefMessageQBA);
@@ -219,25 +181,19 @@ QNearFieldTarget::RequestId QNearFieldTargetPrivateImpl::sendCommand(const QByte
         return requestId;
     }
 
-    // Making QByteArray
-    QByteArray ba(command);
-    jbyteArray jba = env->NewByteArray(ba.size());
-    env->SetByteArrayRegion(jba, 0, ba.size(), reinterpret_cast<jbyte*>(ba.data()));
-
     // Writing
-    QJniObject myNewVal = tagTech.callObjectMethod("transceive", "([B)[B", jba);
+    const QJniArray myNewVal = tagTech.callMethod<jbyte[]>("transceive", command);
     if (!myNewVal.isValid()) {
         // Some devices (Samsung, Huawei) throw an exception when the card is lost:
         // "android.nfc.TagLostException: Tag was lost". But there seems to be a bug that
-        // isConnected still reports true. So we need to invalidate the target to allow
-        // checkIsTargetLost to recognize this missing card.
-        targetIntent = QJniObject();
+        // isConnected still reports true. So we need to invalidate the target as soon as
+        // possible and treat the card as lost.
+        handleTargetLost();
 
         reportError(QNearFieldTarget::CommandError, requestId);
         return requestId;
     }
-    QByteArray result = jbyteArrayToQByteArray(myNewVal.object<jbyteArray>());
-    env->DeleteLocalRef(jba);
+    QByteArray result = myNewVal.toContainer();
 
     setResponseForRequest(requestId, result, false);
 
@@ -277,17 +233,15 @@ QNearFieldTarget::RequestId QNearFieldTargetPrivateImpl::writeNdefMessages(const
 
     // Making NdefMessage object
     const QNdefMessage &message = messages.first();
-    QByteArray ba = message.toByteArray();
-    QJniObject jba = env->NewByteArray(ba.size());
-    env->SetByteArrayRegion(jba.object<jbyteArray>(), 0, ba.size(), reinterpret_cast<jbyte*>(ba.data()));
-    QJniObject jmessage = QJniObject("android/nfc/NdefMessage", "([B)V", jba.object<jbyteArray>());
+    QtJniTypes::NdefMessage jmessage(message.toByteArray());
     if (!jmessage.isValid()) {
         reportError(QNearFieldTarget::UnknownError, requestId);
         return requestId;
     }
 
     // Writing
-    auto methodId = env.findMethod(tagTech.objectClass(), writeMethod, "(Landroid/nfc/NdefMessage;)V");
+    auto methodId =
+            env.findMethod<void, QtJniTypes::NdefMessage>(tagTech.objectClass(), writeMethod);
     if (methodId)
         env->CallVoidMethod(tagTech.object(), methodId, jmessage.object<jobject>());
     if (!methodId || env.checkAndClearExceptions()) {
@@ -325,7 +279,7 @@ void QNearFieldTargetPrivateImpl::checkIsTargetLost()
 
     QJniEnvironment env;
     bool connected = false;
-    auto methodId = env.findMethod(tagTech.objectClass(), "isConnected", "()Z");
+    auto methodId = env.findMethod<jboolean>(tagTech.objectClass(), "isConnected");
     if (methodId)
         connected = env->CallBooleanMethod(tagTech.object(), methodId);
     if (!methodId || env.checkAndClearExceptions()) {
@@ -336,17 +290,17 @@ void QNearFieldTargetPrivateImpl::checkIsTargetLost()
     if (connected)
         return;
 
-    methodId = env.findMethod(tagTech.objectClass(), "connect", "()V");
+    methodId = env.findMethod<void>(tagTech.objectClass(), "connect");
     if (methodId)
         env->CallVoidMethod(tagTech.object(), methodId);
-    if (!methodId || env.checkAndClearExceptions()) {
+    if (!methodId || env.checkAndClearExceptions(QJniEnvironment::OutputMode::Silent)) {
         handleTargetLost();
         return;
     }
-    methodId = env.findMethod(tagTech.objectClass(), "close", "()V");
+    methodId = env.findMethod<void>(tagTech.objectClass(), "close");
     if (methodId)
         env->CallVoidMethod(tagTech.object(), methodId);
-    if (!methodId || env.checkAndClearExceptions())
+    if (!methodId || env.checkAndClearExceptions(QJniEnvironment::OutputMode::Silent))
         handleTargetLost();
 }
 
@@ -364,10 +318,10 @@ void QNearFieldTargetPrivateImpl::updateTechList()
 
     // Getting tech list
     QJniEnvironment env;
-    QJniObject tag = AndroidNfc::getTag(targetIntent);
+    QJniObject tag = QtNfc::getTag(targetIntent);
     Q_ASSERT_X(tag.isValid(), "updateTechList", "could not get Tag object");
 
-    QJniObject techListArray = tag.callObjectMethod("getTechList", "()[Ljava/lang/String;");
+    const QJniArray techListArray = tag.callMethod<QtJniTypes::String[]>("getTechList");
     if (!techListArray.isValid()) {
         handleTargetLost();
         return;
@@ -375,11 +329,8 @@ void QNearFieldTargetPrivateImpl::updateTechList()
 
     // Converting tech list array to QStringList.
     techList.clear();
-    jsize techCount = env->GetArrayLength(techListArray.object<jobjectArray>());
-    for (jsize i = 0; i < techCount; ++i) {
-        QJniObject tech = env->GetObjectArrayElement(techListArray.object<jobjectArray>(), i);
-        techList.append(tech.callObjectMethod<jstring>("toString").toString());
-    }
+    for (const auto &tech : techListArray)
+        techList.append(tech.toString());
 }
 
 void QNearFieldTargetPrivateImpl::updateType()
@@ -391,7 +342,7 @@ QNearFieldTarget::Type QNearFieldTargetPrivateImpl::getTagType() const
 {
     if (techList.contains(NDEFTECHNOLOGY)) {
         QJniObject ndef = getTagTechnology(NDEFTECHNOLOGY);
-        QString qtype = ndef.callObjectMethod("getType", "()Ljava/lang/String;").toString();
+        QString qtype = ndef.callMethod<jstring>("getType").toString();
 
         if (qtype.compare(MIFARETAG) == 0)
             return QNearFieldTarget::MifareTag;
@@ -411,8 +362,7 @@ QNearFieldTarget::Type QNearFieldTargetPrivateImpl::getTagType() const
         // Checking ATQA/SENS_RES
         // xxx0 0000  xxxx xxxx: Identifies tag Type 1 platform
         QJniObject nfca = getTagTechnology(NFCATECHNOLOGY);
-        QJniObject atqaBA = nfca.callObjectMethod("getAtqa", "()[B");
-        QByteArray atqaQBA = jbyteArrayToQByteArray(atqaBA.object<jbyteArray>());
+        const QByteArray atqaQBA = nfca.callMethod<QByteArray>("getAtqa");
         if (atqaQBA.isEmpty())
             return QNearFieldTarget::ProprietaryTag;
         if ((atqaQBA[0] & 0x1F) == 0x00)
@@ -456,7 +406,7 @@ QJniObject QNearFieldTargetPrivateImpl::getTagTechnology(const QString &tech) co
     techClass.replace(QLatin1Char('.'), QLatin1Char('/'));
 
     // Getting requested technology
-    QJniObject tag = AndroidNfc::getTag(targetIntent);
+    QJniObject tag = QtNfc::getTag(targetIntent);
     Q_ASSERT_X(tag.isValid(), "getTagTechnology", "could not get Tag object");
 
     const QString sig = QString::fromUtf8("(Landroid/nfc/Tag;)L%1;");
@@ -488,7 +438,7 @@ bool QNearFieldTargetPrivateImpl::connect()
         return false;
 
     QJniEnvironment env;
-    auto methodId = env.findMethod(tagTech.objectClass(), "isConnected", "()Z");
+    auto methodId = env.findMethod<jboolean>(tagTech.objectClass(), "isConnected");
     bool connected = false;
     if (methodId)
         connected = env->CallBooleanMethod(tagTech.object(), methodId);
@@ -499,7 +449,7 @@ bool QNearFieldTargetPrivateImpl::connect()
         return true;
 
     setCommandTimeout(2000);
-    methodId = env.findMethod(tagTech.objectClass(), "connect", "()V");
+    methodId = env.findMethod<void>(tagTech.objectClass(), "connect");
     if (!methodId)
         return false;
     env->CallVoidMethod(tagTech.object(), methodId);
@@ -512,18 +462,8 @@ bool QNearFieldTargetPrivateImpl::setCommandTimeout(int timeout)
         return false;
 
     QJniEnvironment env;
-    auto methodId = env.findMethod(tagTech.objectClass(), "setTimeout", "(I)V");
+    auto methodId = env.findMethod<void, jint>(tagTech.objectClass(), "setTimeout");
     if (methodId)
         env->CallVoidMethod(tagTech.object(), methodId, timeout);
     return methodId && !env.checkAndClearExceptions();
 }
-
-QByteArray QNearFieldTargetPrivateImpl::jbyteArrayToQByteArray(const jbyteArray &byteArray) const
-{
-    QJniEnvironment env;
-    QByteArray resultArray;
-    jsize len = env->GetArrayLength(byteArray);
-    resultArray.resize(len);
-    env->GetByteArrayRegion(byteArray, 0, len, reinterpret_cast<jbyte*>(resultArray.data()));
-    return resultArray;
-}
